/*
 *$Id: module_main.c,v 1.13 2001/12/01 22:58:57 chooga Exp $
 */

#include <linux/module.h>
#include <linux/version.h>
#include <linux/modversions.h>
#include <linux/config.h>
#include <linux/unistd.h>
#include <linux/sys.h>
#include <linux/errno.h>
#include <linux/kernel.h>
#include <linux/slab.h>
#include <linux/init.h>
#include <linux/sysctl.h>
#include <linux/limits.h>
#include <asm/uaccess.h>
#include <asm/semaphore.h>

#include "filter_vars_public.h"
#include "rule_engine_public.h"
#include "syscall_hijack_public.h"
#include "module_interface.h"
#include "device_file.h"
#include "sct_compat.h"

MODULE_AUTHOR("guy keren <choo@actcom.co.il>, mulix <mulix@actcom.co.il>");

MODULE_DESCRIPTION("System Call Filter Evaluation and Userspace Interface, "
		   "for 'syscalltrack'");

MODULE_LICENSE("GPL");


/*
 * Forward declarations.
 */
int tracker_sysctl(ctl_table *table, int *name, int namelen,
                   void *oldval, size_t *oldlenp,
                   void *newval, size_t newlen, void **context);
void
init_mux_function(int syscall_id, int mux_func_id, func_ptr track_mux_func);

/*
 * Static data structures.
 */

/* array of addresses of original system calls. */
static func_ptr orig_syscalls[NR_syscalls];

/* array of counters of rules registered per syscall.   */
/* used for reference counting, for hijacking purposes. */
static unsigned int src[NR_syscalls];

/* array of rules list, per syscall. */
static rule_list_t* rules[NR_syscalls];


/* structure used to store multiplexing data. */
struct mux_data {
    func_ptr mux_tracker;	/* tracker for the syscall itself.      */
    int num_muxed;		/* number of muxed functions.           */
    func_ptr* muxed_functions;	/* trackers for the muxed functions.    */
    func_ptr* demux_functions;	/* functions allowing re-calling the    */
				/* orig syscall from the function       */
				/* trackers.                            */
    rule_list_t** rules;	/* rules list, once per muxed function. */
} mux_data[NR_syscalls];

/* extract native syscall if from a syscall id. */
#define	SYSCALL_PART_OF_SYSCALL_ID(syscall_id)	((syscall_id) & 0xffff)
/* extract multiplexed function id from a syscall id. */
#define	FUNC_PART_OF_SYSCALL_ID(syscall_id)	(((syscall_id) >> 16) & 0xffff)

/* structure used to store info about type data for a struct type. */
struct struct_data {
    long* offsets_vec;		/* offsets of struct's fields.		*/
    long num_fields;		/* number of fields in struct.		*/
};

static DECLARE_MUTEX(tracker_sem);

/* data used to register sysctl functions for user-mode access to the module. */
static ctl_table tracker_table[] = {
	{CTL_TRACKER_CMD_ADD_RULE,
	 NULL,
	 NULL,
	 1024,
	 0600,
	 NULL,
	 NULL,
	 &tracker_sysctl},
	{CTL_TRACKER_CMD_DEL_RULE, NULL, NULL, 1024,
	 0600, NULL, NULL, &tracker_sysctl},
	{CTL_TRACKER_CMD_ENABLE_RULE, NULL, NULL, 1024,
	 0600, NULL, NULL, &tracker_sysctl},
	{CTL_TRACKER_CMD_DISABLE_RULE, NULL, NULL, 1024,
	 0600, NULL, NULL, &tracker_sysctl},
	{CTL_TRACKER_CMD_DEL_ALL_RULES_FOR_SYSCALL, NULL, NULL, 1024,
	 0600, NULL, NULL, &tracker_sysctl},
	{CTL_TRACKER_CMD_DEL_ALL_MODULE_RULES, NULL, NULL, 1024,
	 0600, NULL, NULL, &tracker_sysctl},
	{CTL_TRACKER_CMD_PRINT_ALL_RULES, NULL, NULL, 1024, 
	 0600, NULL, NULL, &tracker_sysctl}, 
	{0}
};

static ctl_table tracker_top_table[] = {
	{CTL_TRACKER, NULL, NULL, 0, 0500, tracker_table},
	{0}
};


static struct ctl_table_header* tracker_sysctl_header = NULL;

#ifdef MODULE

/* get PAGE_SIZE memory (to copy a string into it, in our case) */
#define __getstring()     (char *)kmalloc(PAGE_SIZE, GFP_KERNEL) 
#define putstring(name)   kfree((name))

static inline int do_getstring(const char* str, char* page)
{
	int retval;
	unsigned long len = PATH_MAX + 1;

	/* TODO: what the hell does this code do? */
	if ((unsigned long) str >= TASK_SIZE) {
		if (!segment_eq(get_fs(), KERNEL_DS))
			return -EFAULT;
	} else if (TASK_SIZE - (unsigned long) str < PATH_MAX + 1)
		len = TASK_SIZE - (unsigned long) str;

	retval = strncpy_from_user((char *)page, str, len);
	if (retval > 0) {
		if (retval < len)
			return 0;
		return -ENAMETOOLONG;
	} else if (!retval) {
	        return -ENOENT;
	}

	return retval;
}

int getstring(const char* str, char** kern_str)
{
	int result;
	char *tmp;

	result = -ENOMEM;
	tmp = __getstring();
	if (tmp)  {
		int retval = do_getstring(str, tmp);
		result = retval;
                *kern_str = tmp;
                if (retval < 0) {
                        putstring(tmp);
			*kern_str = NULL;
                        result = retval;
                }
        }

        return result;
}

filter_node_t* string_param_callback(var_type_t type, int param_index, void* data) {
    filter_node_t *f = NULL;
    char *string;
    int retval;
    
    retval = getstring((char*)data, &string);
    if (retval == 0) {
	f = make_string_val_without_copy(string);
    }
    
    return f;
}

#include "syscall_tracker_autogen.c"


/* adds a new rule for a given system call. */
int
tracker_sysctl_add_rule(int rule_id,
			int syscall_id,
			int mux_func_id,
			filter_node_t* f,
			action_t* a)
{
    int rc;
    int retval = 0;
    int release_sem = 0;
    func_ptr old_func;
    
    printk(KERN_NOTICE "adding tracking rule %d for syscall %d, mux_func_id %d\n",
	   rule_id, syscall_id, mux_func_id);
    /* lock to avoid races with other invocations. */
    if (down_interruptible(&tracker_sem)) {
        retval = -EINTR;
        goto err_add_rule_lock;
    }
    release_sem = 1;
    if (mux_func_id == 0)
        rc = add_tracking_rule(rules[syscall_id], rule_id, f, a);
    else
        rc = add_tracking_rule(mux_data[syscall_id].rules[mux_func_id],
			       rule_id, f, a);
    if (rc < 0) {
	printk (KERN_NOTICE "add tracking_rule returned '%d'\n", rc);
        retval = -ENOMEM; /* todo: something better? */
        goto err_add_rule_add_rule;
    }
    /* make sure we don't release 'f', since it was consumed */
    /* by the 'add_tracking_rule' function.                  */
    f = NULL;
    free_action(a);
    a = NULL;
    if (mux_func_id == 0)
        print_tracking_rules(rules[syscall_id]);
    else
        print_tracking_rules(mux_data[syscall_id].rules[mux_func_id]);

    /* increase reference count. hijack if first rule for syscall. */
    if (src[syscall_id]++ == 0) {
	printk(KERN_NOTICE "first rule for syscall - hijacking syscall\n");
        rc = hijack_syscall(syscall_id,
			    (func_ptr)track_functions[syscall_id],
			    &old_func);
        printk(KERN_NOTICE "hijack_syscall returned '%d'\n", rc);
        if (rc < 0) {
            printk(KERN_NOTICE "hijack_syscall failed, error '%d'\n", rc); 
            retval = (rc != -ERESTARTSYS ? rc : -EINTR);
	    /* remove rule. reduce ref. count. */
	    if (mux_func_id == 0)
    	        del_tracking_rule(rules[syscall_id], rule_id);
	    else
    	        del_tracking_rule(mux_data[syscall_id].rules[mux_func_id],
				  rule_id);
            src[syscall_id]--;
            goto err_add_rule_hijack;
        }
	
        orig_syscalls[syscall_id] = old_func;
        MOD_INC_USE_COUNT;
    }
    /* all went well. */
    goto ret_ok_add_rule;

  err_add_rule_hijack:
  err_add_rule_add_rule:
  err_add_rule_lock:
  ret_ok_add_rule:
    if (release_sem)
	up(&tracker_sem);
    if (f)
	free_filter_node(f);
    if (a)
	free_action(a);

    return retval;
}

/* deletes an existing rule for a given system call. */
int
tracker_sysctl_del_rule(int rule_id, int syscall_id, int mux_func_id)
{
    int rc;
    int retval = 0;
    int release_sem = 0;

    printk(KERN_NOTICE "tracker_sysctl: deleting tracking rule %d for syscall %d\n",
           rule_id, syscall_id);

    /* lock to avoid races with other invocations. */
    if (down_interruptible(&tracker_sem)) {
        retval = -ERESTARTSYS;
        goto err_del_rule_lock;
    }
    release_sem = 1;

    if (mux_func_id == 0)
    	rc = has_tracking_rule(rules[syscall_id], rule_id);
    else
    	rc = has_tracking_rule(mux_data[syscall_id].rules[mux_func_id],
				rule_id);
    if (!rc) {
	printk(KERN_NOTICE "tracker_sysctl: rule '%d' not found.\n", rule_id);
	goto err_del_rule_no_such_rule;
    }

    if (mux_func_id == 0) {
    	del_tracking_rule(rules[syscall_id], rule_id);
    	print_tracking_rules(rules[syscall_id]);
    }
    else {
    	del_tracking_rule(mux_data[syscall_id].rules[mux_func_id], rule_id);
    	print_tracking_rules(mux_data[syscall_id].rules[mux_func_id]);
    }

    if (--src[syscall_id] == 0) {
	printk(KERN_NOTICE "last rule for syscall - releasing syscall\n");
        rc = release_syscall(syscall_id);
        if (rc < 0) {
            printk(KERN_NOTICE "release_syscall failed, error '%d'\n",
		   rc); 
            retval = (rc != -ERESTARTSYS ? rc : -EINTR);
            goto err_del_rule_release;
        }

        orig_syscalls[syscall_id] = (func_ptr)NULL;
        MOD_DEC_USE_COUNT;
    }
    /* all went well. */
    goto ret_ok_del_rule;

  err_del_rule_no_such_rule:
  err_del_rule_release:
  err_del_rule_lock:
  ret_ok_del_rule:
    if (release_sem)
	up(&tracker_sem);

    return retval;
}

/* deletes all existing rules for a given system call. */
/* TODO - handle muxed functions as if they were real syscalls. */
int
tracker_sysctl_del_all_rules_for_syscall(int syscall_id, int mux_func_id)
{
    int rc;
    int retval = 0;
    int release_sem = 0;

    printk(KERN_NOTICE "tracker_sysctl: deleting all tracking rules for syscall %d\n",
           syscall_id);

    /* lock to avoid races with other invocations. */
    if (down_interruptible(&tracker_sem)) {
        retval = -ERESTARTSYS;
        goto err_del_all_rules_for_syscall_lock;
    }
    release_sem = 1;

    if (rules[syscall_id]){
	del_all_tracking_rules_in_list(rules[syscall_id]);
	print_tracking_rules(rules[syscall_id]);
    }
    printk(KERN_NOTICE "tracker_sysctl: del all rules for syscall - releasing syscall\n");
    rc = release_syscall(syscall_id);
    if (rc < 0) {
        printk(KERN_NOTICE "tracker_sysctl: release_syscall failed, error '%d'\n", rc); 
        retval = (rc != -ERESTARTSYS ? rc : -EINTR);
        goto err_del_all_rules_for_syscall_release;
    }
    orig_syscalls[syscall_id] = (func_ptr)NULL;
    src[syscall_id] = 0;

    MOD_DEC_USE_COUNT;

    goto ret_ok_del_all_rules_for_syscall_lock;

  err_del_all_rules_for_syscall_release:
  err_del_all_rules_for_syscall_lock:
  ret_ok_del_all_rules_for_syscall_lock:
    if (release_sem)
	up(&tracker_sem);

    return retval;
}

/* deletes all existing rules in the module */
int
tracker_sysctl_del_all_module_rules(void)
{
    int rc;
    unsigned int i;
    int retval = 0;
    int release_sem = 0;

    printk(KERN_NOTICE "tracker_sysctl: deleting all module tracking rules\n");

    /* lock to avoid races with other invocations. */
    if (down_interruptible(&tracker_sem)) {
        retval = -ERESTARTSYS;
        goto err_del_all_module_rules_lock;
    }
    release_sem = 1;

    /* iterate over all syscalls and release the hijacked ones */
    for (i = 0; i < NR_syscalls; ++i){
	/* not hijacked? go on */
	if (!orig_syscalls[i]) 
	    continue;
	
	if (rules[i])
	    del_all_tracking_rules_in_list(rules[i]);
	/* ahndle muxed syscalls. */
	if (mux_data[i].num_muxed > 0 && mux_data[i].rules) {
	    int j;
	    for (j=0; j < mux_data[i].num_muxed; j++) {
		if (mux_data[i].rules[j])
		    del_all_tracking_rules_in_list(mux_data[i].rules[j]);
	    }
	}
	
	rc = release_syscall(i);
	if (rc < 0) {
	    printk(KERN_NOTICE "tracker_sysctl: release_syscall for syscall '%d' failed, "
		   "error '%d'\n", i, rc); 
	    retval = (rc != -ERESTARTSYS ? rc : -EINTR);
	    goto err_del_all_module_rules_release;
	}
	orig_syscalls[i] = (func_ptr)NULL;
	src[i] = 0;
	MOD_DEC_USE_COUNT;
    }

    goto ret_ok_del_all_module_rules;

  err_del_all_module_rules_lock:
  err_del_all_module_rules_release:
  ret_ok_del_all_module_rules:
    if (release_sem)
	up(&tracker_sem);

   return retval;
}

/* prints all existing rules in the module */
int
tracker_sysctl_print_all_rules(void)
{
    unsigned int i;
    int retval = 0;
    int printed = 0;
    
    /* lock to avoid races with other invocations. */
    if (down_interruptible(&tracker_sem)) {
        retval = -ERESTARTSYS;
        goto err_print_module_rules_lock;
    }
    
    printk (KERN_NOTICE "**************************************\n");
    printk (KERN_NOTICE "syscalltrack - printing defined rules\n");
    for (i = 0; i < NR_syscalls; ++i)
    {
	if (src[i]){
	    if (rules[i]) {
	        printk(KERN_NOTICE "rules for syscall '%d':\n", i);
	        print_tracking_rules(rules[i]);
	        printed = 1;
	    }
	    /* handled muxed syscalls. */
	    if (mux_data[i].num_muxed > 0 && mux_data[i].rules) {
		int j;
		for (j=0; j<mux_data[i].num_muxed; j++) {
	    	    if (mux_data[i].rules[j]) {
	        	printk(KERN_NOTICE "rules for syscall '%d:%d':\n", i,j);
	        	print_tracking_rules(mux_data[i].rules[j]);
	        	printed = 1;
	    	    }
		}
	    }
	}
    }
    if (!printed){
	printk (KERN_NOTICE "no tracking rules registered currently\n");
    }
    printk (KERN_NOTICE "**************************************\n");
    
    up(&tracker_sem);

 err_print_module_rules_lock:
    return retval;
}

/* sysctl handling routine. allows adding/deleting/enabling/disabling rules. */
int tracker_sysctl(ctl_table *table,
		   int *name, int namelen,
                   void *oldval, size_t *oldlenp,
                   void *newval, size_t newlen, void **context)
{
    int i;
    int rule_id;
    int syscall_id;
    action_t* a;
    filter_node_t* f;
    int rc, retval = 0;
    unsigned char* buf = NULL;
    int mux_func_id = 0;

    printk(KERN_DEBUG "tracker_sysctl: namelen = '%d', newlen = '%d'\n", namelen, newlen);
    for (i=0; i<namelen; i++) {
	printk(KERN_DEBUG "tracker_sysctl: name[%d] = %d\n", i, name[i]);
    }
    if (newlen <= 0 || namelen != 1) {
	retval = -EINVAL;
	goto err_out;
    }

    buf = kmalloc(newlen, GFP_KERNEL);
    if (!buf){
	retval = -ENOMEM;
	printk(KERN_NOTICE "tracker_sysctl: buffer allocation failed\n");
	goto err_oom;
    }
    if (copy_from_user(buf, newval, newlen)) {
	retval = -EFAULT; /* TODO: enomem? eaccess? */
	printk(KERN_NOTICE "tracker_sysctl: failed to copy from user\n");
	goto err_copy_from_user;
    }
    printk(KERN_NOTICE "tracker_sysctl: copied data from user\n");
    rc = deserialize_tracking_rule(buf, newlen, &rule_id, &syscall_id, &a, &f);
    if (rc <= 0) {
	printk(KERN_NOTICE "tracker_sysctl: failed to deserialize tracking rule, "
	       "err '%d'\n", rc);
	retval = -EINVAL; /* todo: anything better? */
	goto err_deserialize;
    }

    mux_func_id = FUNC_PART_OF_SYSCALL_ID(syscall_id);
    syscall_id = SYSCALL_PART_OF_SYSCALL_ID(syscall_id);

    /* sanity checks */
    if (syscall_id < 0 || syscall_id > NR_syscalls) {
        printk(KERN_NOTICE "syscall id %d out of range\n",
	       syscall_id);
	goto err_invalid_syscall_id;
    }
    if (mux_func_id < 0 ||
	(mux_func_id > 0 && mux_func_id >= mux_data[syscall_id].num_muxed)) {
        printk(KERN_NOTICE "syscall id %d , mux function id %d out of range\n",
	       syscall_id, mux_func_id);
	goto err_invalid_syscall_id;
    }
    if (mux_func_id > 0 &&
	(!mux_data[syscall_id].rules || !mux_data[syscall_id].rules[mux_func_id])) {
        printk(KERN_NOTICE "syscall id %d , mux function id %d not supported\n",
	       syscall_id, mux_func_id);
	goto err_invalid_syscall_id;
    }

    /* which operation did the user request? */
    switch (name[0]) {
        case(CTL_TRACKER_CMD_ADD_RULE):
	    retval = tracker_sysctl_add_rule(rule_id, syscall_id, mux_func_id,
					     f, a);
	    break;
	
        case(CTL_TRACKER_CMD_DEL_RULE):
	    retval = tracker_sysctl_del_rule(rule_id, syscall_id, mux_func_id);
	    break;
	
	case(CTL_TRACKER_CMD_DEL_ALL_RULES_FOR_SYSCALL):
	    retval = tracker_sysctl_del_all_rules_for_syscall(syscall_id,
							      mux_func_id);
	    break;

        case(CTL_TRACKER_CMD_DEL_ALL_MODULE_RULES):
	    retval = tracker_sysctl_del_all_module_rules();
 	    break;

        case(CTL_TRACKER_CMD_PRINT_ALL_RULES):
	    retval = tracker_sysctl_print_all_rules();
	    break;

        default:
	    printk(KERN_NOTICE "tracker_sysctl: invalid or unsupported sysctl "
		   "operation %d\n", name[0]);
	    retval = -EINVAL;
    }
    /* all is well - we can return peacefully. */
    goto ret_out;

 err_invalid_syscall_id:
 err_deserialize:
 err_copy_from_user:
 err_oom:
 err_out:
 ret_out:
    if (buf)
        kfree(buf);

    return retval;
}

/* init muxed syscalls data, for the given syscall id. */
void
init_mux_syscall(int syscall_id,
		 func_ptr syscall_track_func,
		 int num_muxed_functions)
{
    int j;

    track_functions[syscall_id] = syscall_track_func;
    mux_data[syscall_id].mux_tracker = syscall_track_func;
    mux_data[syscall_id].num_muxed = num_muxed_functions;
    mux_data[syscall_id].muxed_functions =
    	(func_ptr*)kmalloc(num_muxed_functions*sizeof(func_ptr), GFP_KERNEL);
    mux_data[syscall_id].rules =
    	(rule_list_t**)kmalloc(num_muxed_functions*sizeof(rule_list_t*),
			       GFP_KERNEL);
    for (j=0; j < num_muxed_functions; j++) {
        mux_data[syscall_id].muxed_functions[j] = (func_ptr)NULL;
        mux_data[syscall_id].rules[j] = (rule_list_t*)NULL;
    }
}

void
init_mux_function(int syscall_id, int mux_func_id, func_ptr track_mux_func)
{
    mux_data[syscall_id].muxed_functions[mux_func_id] = track_mux_func;
    mux_data[syscall_id].rules[mux_func_id] = create_tracking_rule_list();
}

#if 0  /* UNUSED for now */
/* forward referenec for a special function to handle 'socketcall' syscall. */
extern asmlinkage long sct_sys_socketcall(int call, unsigned long *args);
#endif /* 0 */

int
init_module (void)
{
	int i, result;

	for (i=0; i<NR_syscalls; ++i) {
		orig_syscalls[i] = (func_ptr)NULL;
		src[i] = 0;
		rules[i] = create_tracking_rule_list();
		mux_data[i].mux_tracker = (func_ptr)NULL;
		mux_data[i].num_muxed = 0;
		mux_data[i].muxed_functions = (func_ptr*)NULL;
		mux_data[i].demux_functions = (func_ptr*)NULL;
		mux_data[i].rules = (rule_list_t**)NULL;
	}

#if 0 /* UNUSED for now */
	/* special handling for 'socketcall' syscall - */
	/* register a special tracker function.        */
	init_mux_syscall(102, (func_ptr)sct_sys_socketcall, 18);
#endif /* 0 */

	init_mux_functions();
		
	/* init calculation of data used to handle struct syscall parameters. */
	init_struct_offsets();

	/* register sysctl handling function. */
 	tracker_sysctl_header = register_sysctl_table(tracker_top_table, 0);
	printk(KERN_DEBUG "tracker_sysctl_header - %p\n", tracker_sysctl_header); 

	/* register device file handling functions */
	SET_MODULE_OWNER(&sct_fops);

	/*
	 * Register your major, and accept a dynamic number
	 */
	result = register_chrdev(sct_major, "sct_rules", &sct_fops);
	if (result < 0) {
	    if (tracker_sysctl_header)
		unregister_sysctl_table(tracker_sysctl_header);
	    return result;
	}
	
	if (sct_major == 0) 
	    sct_major = result; /* dynamic */
	
	sema_init(&sct_dev.sem, 1);

	return 0;
}

void
cleanup_module (void)
{
	int i;

        printk(KERN_NOTICE "sct_rules.o: cleaning up syscall filter module\n");

	/* register sysctl handling function. */
	if (tracker_sysctl_header)
		unregister_sysctl_table(tracker_sysctl_header);

	for (i=0; i<NR_syscalls; ++i) {
		delete_tracking_rule_list(rules[i]);
		mux_data[i].mux_tracker = (func_ptr)NULL;
		mux_data[i].num_muxed = 0;
		if (mux_data[i].muxed_functions)
		    kfree(mux_data[i].muxed_functions);
		if (mux_data[i].demux_functions)
		    kfree(mux_data[i].demux_functions);
		if (mux_data[i].rules) {
		    int j;
		    for (j=0; j<mux_data[i].num_muxed; j++)
		        delete_tracking_rule_list(mux_data[i].rules[j]);
		    kfree(mux_data[i].rules);
		}
	}

	unregister_chrdev(sct_major, "sct_rules");
}



/***********************************************************************/    
/***********************************************************************/    

/* special handlign for the 'socketcall' syscall, and its sub-elements. */

#if 0 /* UNUSED for now */
/* special function, used to track 'socketcall'. */
/* Argument list sizes for sys_socketcall */
#define AL(x) ((x) * sizeof(unsigned long))
static unsigned char nargs[18]={AL(0),AL(3),AL(3),AL(3),AL(2),AL(3),
				AL(3),AL(3),AL(4),AL(4),AL(4),AL(6),
				AL(6),AL(2),AL(5),AL(5),AL(3),AL(3)};
#undef AL
#endif /* 0 */


/*
 *	System call vectors. 
 *
 *	Argument checking cleaned up. Saved 20% in size.
 *  This function doesn't need to set the kernel lock because
 *  it is set by the callees. 
 */

#if 0
asmlinkage long sct_sys_socketpair(int family, int type, int protocol, int usockvec[2])
asmlinkage long sct_sys_bind(int fd, struct sockaddr *umyaddr, int addrlen)
asmlinkage long sct_sys_listen(int fd, int backlog)
asmlinkage long sct_sys_accept(int fd, struct sockaddr *upeer_sockaddr, int *upeer_addrlen)
asmlinkage long sct_sys_connect(int fd, struct sockaddr *uservaddr, int addrlen)
asmlinkage long sct_sys_getsockname(int fd, struct sockaddr *usockaddr, int *usockaddr_len)
asmlinkage long sct_sys_getpeername(int fd, struct sockaddr *usockaddr, int *usockaddr_len)
asmlinkage long sct_sys_sendto(int fd, void * buff, size_t len, unsigned flags,
			   struct sockaddr *addr, int addr_len)
asmlinkage long sct_sys_send(int fd, void * buff, size_t len, unsigned flags)
asmlinkage long sct_sys_recvfrom(int fd, void * ubuf, size_t size, unsigned flags,
			     struct sockaddr *addr, int *addr_len)
asmlinkage long sct_sys_recv(int fd, void * ubuf, size_t size, unsigned flags)
asmlinkage long sct_sys_setsockopt(int fd, int level, int optname, char *optval, int optlen)
asmlinkage long sct_sys_getsockopt(int fd, int level, int optname, char *optval, int *optlen)
asmlinkage long sct_sys_shutdown(int fd, int how)
asmlinkage long sct_sys_sendmsg(int fd, struct msghdr *msg, unsigned flags)
asmlinkage long sct_sys_recvmsg(int fd, struct msghdr *msg, unsigned int flags)
#endif /* 0 */


#if 0 /* UNUSED for now */
asmlinkage long sct_sys_socketcall(int call, unsigned long *args)
{
        typedef int
                (*hijack_func_102_b)(int call, unsigned long *args);
        hijack_func_102_b func = (hijack_func_102_b)orig_syscalls[102];

        /*printk(KERN_INFO "sct_sys_socketcall - call '%d'\n", call);*/

	unsigned long a[6];
	unsigned long a0,a1;
	int err;

	if(call<1||call>SYS_RECVMSG)
		return -EINVAL;

	/* copy_from_user should be SMP safe. */
	if (copy_from_user(a, args, nargs[call]))
		return -EFAULT;
		
	a0=a[0];
	a1=a[1];
	
	switch(call) 
	{
		case SYS_SOCKET:
			err = track_syscall_102_1(a0,a1,a[2]);
			if (err == 0)
			    return (func ? (*func)(call, args) : -ENOSYS);
			break;
		case SYS_BIND:
			err = track_syscall_102_2(a0,(struct sockaddr *)a1, a[2]);
			if (err == 0)
			    return (func ? (*func)(call, args) : -ENOSYS);
			break;
		case SYS_CONNECT:
			err = track_syscall_102_3(a0, (struct sockaddr *)a1, a[2]);
			if (err == 0)
			    return (func ? (*func)(call, args) : -ENOSYS);
			break;
		case SYS_LISTEN:
			err = track_syscall_102_4(a0,a1);
			if (err == 0)
			    return (func ? (*func)(call, args) : -ENOSYS);
			break;
		case SYS_ACCEPT:
			err = track_syscall_102_5(a0,(struct sockaddr *)a1, (int *)a[2]);
			if (err == 0)
			    return (func ? (*func)(call, args) : -ENOSYS);
			break;
		case SYS_SHUTDOWN:
			err = track_syscall_102_13(a0,a1);
			if (err == 0)
			    return (func ? (*func)(call, args) : -ENOSYS);
			break;
		default:
			return (func ? (*func)(call, args) : -ENOSYS);
#if 0
		case SYS_GETSOCKNAME:
			err = demux_sys_getsockname(a0,(struct sockaddr *)a1, (int *)a[2]);
			break;
		case SYS_GETPEERNAME:
			err = demux_sys_getpeername(a0, (struct sockaddr *)a1, (int *)a[2]);
			break;
		case SYS_SOCKETPAIR:
			err = demux_sys_socketpair(a0,a1, a[2], (int *)a[3]);
			break;
		case SYS_SEND:
			err = demux_sys_send(a0, (void *)a1, a[2], a[3]);
			break;
		case SYS_SENDTO:
			err = demux_sys_sendto(a0,(void *)a1, a[2], a[3],
					 (struct sockaddr *)a[4], a[5]);
			break;
		case SYS_RECV:
			err = demux_sys_recv(a0, (void *)a1, a[2], a[3]);
			break;
		case SYS_RECVFROM:
			err = demux_sys_recvfrom(a0, (void *)a1, a[2], a[3],
					   (struct sockaddr *)a[4], (int *)a[5]);
			break;
		case SYS_SETSOCKOPT:
			err = demux_sys_setsockopt(a0, a1, a[2], (char *)a[3], a[4]);
			break;
		case SYS_GETSOCKOPT:
			err = demux_sys_getsockopt(a0, a1, a[2], (char *)a[3], (int *)a[4]);
			break;
		case SYS_SENDMSG:
			err = demux_sys_sendmsg(a0, (struct msghdr *) a1, a[2]);
			break;
		case SYS_RECVMSG:
			err = demux_sys_recvmsg(a0, (struct msghdr *) a1, a[2]);
			break;
		default:
			err = -EINVAL;
			break;
#endif /* 0 */
	}
	return err;
}
#endif /* 0 */

/* end of special section for dealing with socketcall. */
/************************************************************************/
/************************************************************************/

#endif /* MODULE */
