/*
 *$Id: syscall_hijack.c,v 1.5 2001/12/01 14:46:49 chooga Exp $
 */
#include <linux/module.h>
#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/sys.h>
#include <linux/stddef.h>
#include <linux/string.h>
#include <linux/errno.h>
#include <asm/semaphore.h>

#include "syscall_hijack_public.h"
#include "sct_compat.h" 

MODULE_AUTHOR("guy keren <choo@actcom.co.il>, mulix <mulix@actcom.co.il>");

MODULE_DESCRIPTION("System Call Hijacking and Releasing, for 'syscalltrack'");

MODULE_LICENSE("GPL");

/* table of original system calls, for un-hijacking purposes. */
static func_ptr orig_sys_call_table[NR_syscalls];
/* table of hijacking functions. */
static func_ptr hijacking_functions[NR_syscalls];

/* table of semaphores, one per system call. */
static struct semaphore syscall_sem[NR_syscalls];

static int syscall_ref_count[NR_syscalls];

extern func_ptr* sys_call_table[NR_syscalls];

#include "syscall_hijack_autogen.c"

/* hijack the given system call. returns '0' on success,
   '1' if the syscall is already hijacked, and 
   a negative int on error. old_function points to the 
   old function on if the syscall was hijacked, or to NULL otherwise */
int hijack_syscall(int syscall_id, func_ptr function, func_ptr* old_function)
{
    int retval = -ENOSYS;
    int rc;

    if (syscall_id >= 0 && syscall_id < NR_syscalls) {
	printk (KERN_DEBUG "hijack_syscall: %d: before down_interruptible\n", syscall_id);
	rc = down_interruptible(&syscall_sem[syscall_id]);
	if (rc < 0){
	    *old_function = 0;
	    return -ERESTARTSYS;
	}
	if (syscall_ref_count[syscall_id] > 0) {
	    printk (KERN_NOTICE "hijack_syscall: syscall %d busy\n", syscall_id);
	    up(&syscall_sem[syscall_id]);
	    return 1;
	} 
	
	printk (KERN_INFO "hijack_syscall: beginning %d hijack\n", syscall_id);
	if (!stub_functions[syscall_id]) {
	    printk (KERN_NOTICE "hijack_syscall: %d: unsupported syscall\n", syscall_id);
	    retval = -ENOENT;
	}
	else if (!orig_sys_call_table[syscall_id]) {
	    /* func_ptr* sys_call_table_addr = (func_ptr*)sys_call_table;  */
	    func_ptr* sys_call_table_addr = (func_ptr*)
		get_module_symbol(NULL, versioned_string(sys_call_table)); 
	    if (sys_call_table_addr) {
		func_ptr orig_syscall =
		    sys_call_table_addr[syscall_id];

		MOD_INC_USE_COUNT;
		       
		/* put into the syscall table the stub function */
		sys_call_table_addr[syscall_id] =
		    (func_ptr)stub_functions[syscall_id];
		       
		/* put into the hijacking functions table the user supplied function */
		hijacking_functions[syscall_id] = function;
		       
		/* put the original syscall into the orig syscalls table */
		orig_sys_call_table[syscall_id] = orig_syscall;
		       
		/* and return that function */
		*old_function = orig_syscall;
	        retval = 0;
	    }
	    else {
	        printk (KERN_NOTICE "hijack_syscall: %d: already hijacked\n", syscall_id);
	        retval = -ENOENT;
	    }
	}
	else {
	    printk (KERN_NOTICE "hijack_syscall: %d: already hijacked\n", syscall_id);
	    retval = 1;
	}
	printk (KERN_DEBUG "hijack_syscall: %d: before up\n", syscall_id);
	up(&syscall_sem[syscall_id]); 
	printk (KERN_DEBUG "hijack_syscall: %d: after up\n", syscall_id);
    }

    return retval;
}

/* releases a previously hijacked system call. */
int release_syscall(int syscall_id)
{
    int retval = 0; /* assume success, if there's no syscall to release */

    if (syscall_id >= 0 && syscall_id < NR_syscalls) {
	printk (KERN_DEBUG "release_syscall: %d: before down_interruptible\n", syscall_id);
	retval = down_interruptible(&syscall_sem[syscall_id]);
	if (retval < 0)
	    return -ERESTARTSYS;
	if (syscall_ref_count[syscall_id] > 0) {
	    printk (KERN_NOTICE "release_syscall: %d: syscall busy (%d active invocations)\n", syscall_id, syscall_ref_count[syscall_id]);
	    up(&syscall_sem[syscall_id]);
	    return -EBUSY;
	} 

	if (orig_sys_call_table[syscall_id]) {
	    /* func_ptr* sys_call_table_addr = (func_ptr*)sys_call_table;  */
	    func_ptr* sys_call_table_addr = (func_ptr*)
		get_module_symbol(NULL, versioned_string(sys_call_table));		
	    if (sys_call_table_addr) {
		func_ptr orig_syscall =
		    orig_sys_call_table[syscall_id];
		
		sys_call_table_addr[syscall_id] = orig_syscall;
		orig_sys_call_table[syscall_id] = (func_ptr)NULL;
		hijacking_functions[syscall_id] = (func_ptr)NULL;
		MOD_DEC_USE_COUNT;
	    }
	}
	printk (KERN_DEBUG "release_syscall: %d: before up\n", syscall_id);
	up(&syscall_sem[syscall_id]); 
	printk (KERN_DEBUG "release_syscall: %d: after up\n", syscall_id);
    }
    return retval;
}

int
init_module(void)
{
        int i;

	memset(orig_sys_call_table, 0, sizeof(orig_sys_call_table));
	memset(hijacking_functions, 0, sizeof(hijacking_functions));
	memset(syscall_ref_count, 0, sizeof(syscall_ref_count));
	
        for (i=0; i<NR_syscalls; i++) {
	     init_MUTEX(&syscall_sem[i]);
        }

        /* disable unloading this module. */
	/* disabling disabled while debugging */
        /* MOD_INC_USE_COUNT; */

        return 0;
}

void
clean_module(void)
{
        /* do nothing - this module cannot be unloaded. */
        printk(KERN_NOTICE "Error - syscall_hijack module should NOT be unloaded!!");
}


EXPORT_SYMBOL(hijack_syscall);
EXPORT_SYMBOL(release_syscall); 
